import numpy as np
from itertools import compress
from .field import *
from .quasiparticles import *
from scipy.interpolate import RegularGridInterpolator 
from .charge_propagation import *
from tqdm.auto import tqdm
import csv
from .silicon import *
import copy
from .mobility import *

#from line_profiler import LineProfiler

#from joblib import Parallel, delayed
#import multiprocessing

#num_cores = multiprocessing.cpu_count()

class Simulation:
    '''
    Object contains all nessie objects needed to simulate a signal. 
    Currently assumes only one contact.
    
    '''
    def __init__(self, _name, _temp, _electricField=None, _weightingPotential=None, _electricPotential=None,
                 _weightingField=None, _cceField=None, _chargeCaptureField=None, _electronicResponse=None, contacts=1,
                 _impurityConcentration= lambda x, y, z : 1e16, _mobility = [generalized_mobility_el, generalized_mobility_h]):        
        self.name = _name
        self.electricField = _electricField
        self.electricPotential = _electricPotential
        self.weightingPotential = _weightingPotential
        self.weightingField = _weightingField
        self.cceField = _cceField
        self.chargeCaptureField = _chargeCaptureField
        self.electronicResponse = _electronicResponse
        self.temp = _temp
        self.bounds = None
        self.contacts = contacts
        self.impurityConcentration = _impurityConcentration
        self.mobility = _mobility

        if contacts is not None and type(_weightingPotential) is not list:
            centers = find_centers(contacts)
            wps = []
            for i in range(contacts):
                wp_temp = copy.deepcopy(_weightingPotential)
                wp_temp.shift(centers[i]+(0,))
                wps.append(wp_temp)
            self.weightingPotential=wps

        if _electricField is None and _electricPotential is not None:
            self.electricField = _electricPotential.toField()
            
    def __str__(self):
       return ("Nessie Simulation object \n Name: %s \n Electric Field: %s \n Weighting Potential: %s \n CCE Field: %s \n Charge Capture Field: %s \n Electronic Response: %s \n"
                % (self.name, self.electricField,self.weightingPotential, self.cceField, self.chargeCaptureField,          
                   self.electronicResponse))
    
    def setTemp(self,T):
        self.temp = T
        return None
    
    def setMobility(self,mobility_e, mobility_h):
        self.mobility = [mobility_e, mobility_h]
        return None
    
    def setIDP(self, IDP): #IDP(x,y,z)->NI[m^-3]
        if type(IDP) is Potential:
            self.impurityConcentration = IDP.interpolate()

        else:
            self.impurityConcentration = IDP
        return None
        
    def setBounds(self, bounds):
        self.bounds = bounds
        return None
    
    def setElectricField(self):
        grid = self.electricPotential.grid
        gradient = np.gradient(self.electricPotential.data, grid[0],grid[1],grid[2])
        try:
            self.electricField = Field("Electric Field", gradient[0],gradient[1],gradient[2], self.electricPotential.grid)
        except:
            print("No electric potential from which to calculate weighting field.")
        return None
    
    def recenter_hex_contacts(self, R, s):
        centers = find_centers(self.contacts, R, s)
        for i in range(centers):
            self.weightingPotential[i].shift(centers[i]+(0,))
    
    def setWeightingField(self):
        contacts = range(self.contacts)
        try:
            self.weightingField = [self.weightingPotential[contact].toField() for contact in contacts]
        except:
            print("No weighting potential from which to calculate weighting field.")
        return None

    def setChargeCollectionEfficiencyField(self):
        return None

    def setChargeCaptureField(self):
        return None

    def setElectronicResponse(self, spiceFile=None, t1=None,t2=None, length=7000):
        '''
        This either takes spice output csv file or assumes a RC-CR signal shaping and the user has to specify each time constant
        approximate values for Nab are 5 us and 7 ns
        '''
        if spiceFile is not None:
            ts = []
            step = []
            with open(spiceFile,'r') as csvfile:
                plots = csv.reader(csvfile, delimiter=',')
                for row in plots:
                    ts.append(float(row[0])*1e-9) #convert from ns to s
                    step.append(float(row[1]))
            self.electronicResponse = {"times":ts, "step":step}
        
        elif t1 is not None and t2 is not None:
            ts = np.arange(length)*1e-9
            step = 1/(t1-t2) * (np.exp(-ts/t1)-np.exp(-ts/t2))
            step = (np.exp(-ts/t1)-np.exp(-ts/t2))
            self.electronicResponse = {"times":ts, "step":step}

        return None

    def simulate(self, events, ds, dt, coulomb=False, diffusion=False, capture=False, d=None, interp3d = True, maxPairs=100, 
                Efield=None, bounds=None, silence=False):
        '''
        Where it all happens! 
        When calling this function you determine which effects you want to simulate (eg. plasma, diffusion, etc.)
        Reduce maxPairs to run faster by having fewer quasiparticles with greater charge.
        Setting interp3d to true here will use our cython interpolator throughout the simulation.
        For machines with many cores you can try setting parallel=True. We see slight speedups but your mileage may vary. 
        '''
    
        # Get electric field interpolations
        if Efield is None:
            Ex_i, Ey_i, Ez_i, Emag_i = self.electricField.interpolate(interp3d)
        else:
            Ex_i, Ey_i, Ez_i, Emag_i= Efield
        
        if bounds is None:
            simBounds = self.bounds if self.bounds is not None else [[axis[0],axis[-1]] for axis in self.electricField.grid]
        else: 
            simBounds = bounds

        #Find electron and hole drift paths for each event
        for i in (t:=tqdm(range(len(events)))):
            t.set_description(f"Drift Calculation", refresh=True)
            event = events[i]
            #rint will give an integer number of e-h pairs, then adjust for variance using sigma = sqrt(FN)
            # assumes in linear regeme of fano factor
            pairs_0 = event.dE/ephBestFit(self.temp)
            pairs = np.rint(np.random.normal(pairs_0,np.sqrt(Fano_Si*pairs_0))) 

            # electron charge cloud assembly
            cc_e = []
            # hole charge cloud assembly
            cc_h = []

            # Add charge cloud parts at each position where energy was deposited
            for j in tqdm(range(len(event.pos)), disable=silence):
                pairNr = int(pairs[j])
                factor = 1
                if pairNr > maxPairs:
                    factor = pairNr/maxPairs
                    pairNr = maxPairs

                # TODO: Provide a radius to smooth initial charge cloud (currently 0)
                cc_e = cc_e + initializeChargeCloud(-factor*qe_SI, factor*me_SI, pairNr, event.times[j], 0, event.pos[j])
                cc_h = cc_h + initializeChargeCloud(factor*qe_SI, factor*me_SI, pairNr, event.times[j], 0, event.pos[j])

            cc = cc_e + cc_h

            alive = np.ones(len(cc)) == 1
            # print("Total quasiparticles: %d" % len(cc))
            # t.set_description(f"Event {i}, Total quasiparticles: {len(cc)}", refresh=True)
            t.set_postfix_str(f"Event {i}, Total quasiparticles: {len(cc)}", refresh=True)

            # Loop over alive particles until all have been stopped/collected
            pbar = tqdm(disable=silence)
            while np.any(alive):
                cc_new = updateQuasiParticles(list(compress(cc, alive)), ds, dt, Ex_i, Ey_i, Ez_i, Emag_i, simBounds,
                                              self.temp, diffusion=diffusion, coulomb=coulomb,NI=self.impurityConcentration,
                                              mobility_e = self.mobility[0], mobility_h = self.mobility[1])
                
                counter = 0
                for j in range(len(alive)):
                    if alive[j]:
                        cc[j] = cc_new[counter]
                        counter+=1

                alive = np.array([o.alive for o in cc])
                pbar.update(1)

            event.quasiparticles = cc

        return events

    def calculateInducedCurrent(self, events, dt, contacts = None, interp3d=True):
        if contacts is None: contacts=np.arange(self.contacts)
        if self.weightingField is None: self.setWeightingField()
        for contact in contacts:
            weightingFieldx_interp, weightingFieldy_interp, weightingFieldz_interp, weightingFieldMag_interp = self.weightingField[contact].interpolate(interp3d)
            wf_interp = [weightingFieldx_interp, weightingFieldy_interp, weightingFieldz_interp, weightingFieldMag_interp]
        
            for event in events:
                event.calculateInducedCurrent(dt, wf_interp, contact)

    def calculateInducedCharge(self, events, contacts = None):
        if contacts is None: contacts=np.arange(self.contacts)
        for contact in contacts:
            for event in events:
                event.calculateIntegratedCharge(contact)

    def calculateElectronicResponse(self, events, contacts = None):
        if contacts is None: contacts=np.arange(self.contacts)
        for event in events:
            for contact in contacts:
                event.convolveElectronicResponse(self.electronicResponse, contact)

    def getSignals(self, events, dt, contacts = None, interp3d=True):
        self.calculateInducedCurrent(events, dt, contacts, interp3d)
        self.calculateElectronicResponse(events, contacts)

def find_centers(N,R=5.15e-3,s=0.1e-3):    
    ''' This functions finds where to place the centers of hexagonal configuration depending on the hexagon radius 
        and seperation between hexagons. Default values are correct for Nab detectors. This is here primarily to copy
        over weighting potentials for multiple contact sims. '''
    
    centers = [(0,0)]
    o_centers = [(0,0)]
    
    r = (np.sqrt(3)*R+s)
    theta = np.radians(np.arange(0,360,60))
    dx = np.round(r*np.cos(theta), 11)
    dy = np.round(r*np.sin(theta),11)
    
    while len(centers) < N:
        new_centers = []
        for center in o_centers:
            oldx, oldy = center
            for i in range(len(theta)):
                new_center = (np.round(oldx+dx[i], 8), np.round(oldy+dy[i], 8))
                if new_center not in centers:
                    new_centers.append(new_center)
                    centers.append(new_center)
        o_centers = new_centers
    return centers[:N]


import os
import sys
sys.path.append(os.getcwd()+"/src/")
import nesse
import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import RegularGridInterpolator 
import time


def main():
    EF_filename = "config/Fields/NessieEF_4e7Linear0-150V_grid.hf"
    WP_filename = "config/Fields/NessieWP_4e7Linear0-150V_grid.hf"

    Efield=nesse.eFieldFromH5(EF_filename)  
    weightingPotential = nessie.weightingPotentialFromH5(WP_filename)

    x = weightingPotential.grid[0].astype("double")
    y = weightingPotential.grid[1].astype("double")
    z = weightingPotential.grid[2].astype("double")
    
    v = weightingPotential.data.astype("double",order="C")
    
    interp = nesse.Interp3D(v,x,y,z)
    interp_si = RegularGridInterpolator((x,y,z),v)
    
    zs = np.linspace(-0.001,0.003, 1000)
    
    t0 = time.time()
    wp = [interp([0,0,z]) for z in zs]
    t1 = time.time() 
    wp_si = [interp_si([0,0,z])[0] for z in zs]
    t2 = time.time()
    print(t1-t0, t2-t1)
    plt.plot(zs,wp)
    plt.plot(zs,wp_si)
    plt.show()
    
    
if __name__ == "__main__":
    main()



from scipy.interpolate import RegularGridInterpolator 
from .interp_3d import *
import numpy as np
import h5py

class Potential:
    '''
    The Potential class is used to create electric and weighting potential objects. 
    These can then be converted to Field objects for simulation. 
    We use Potential objects mainly for plotting and transferring field information between nessie and our other field solving tools. 
    data is assumed to be a 3-dimensional array. grid is an array of axes (x,y,z) where each axis is an array of positions in meters. 
    '''
    def __init__(self, _name, _data=None, _grid=None):
        self.name = _name
        self.data = _data
        self.grid = _grid

    def __str__(self):
        return ("Nessie Potential object \n Name: %s \n Size: %s \n" 
                % (self.name, np.shape(self.data)))
    
    def shift(self, shift=(0,0,0)):
        for i in [0,1,2]:
            self.grid[i] = self.grid[i]+shift[i]

    def toField(self):
        gradient = np.gradient(self.data, self.grid[0],self.grid[1], self.grid[2])
        return Field("Weighting Field", gradient[0],gradient[1],gradient[2],
                                          self.grid)
    
    def saveToH5(self, filename):
        if filename is None: filename = self.name
        f_new = h5py.File(filename + ".hf", "w")
        f_new["wp"] = self.data
        f_new['gridx'] = self.grid[0]
        f_new['gridy'] = self.grid[1]
        f_new['gridz'] = self.grid[2]
        f_new.close()

    def interpolate(self, interp3d=False):
        '''
        Returns linear interpolations of the field. 
        interp3d is a cython implementation of grid interpolation, we have modified this from https://github.com/jglaser/interp3d to work for non-regular grids. 
        If interp3d is set to False instead the scipy RegularGridInterpolator will be used. 
        This is a pure python interpolator and is significantly slower. 
        Each interpolator takes input differently so which is used is tracked throughout the simulation.  
        '''
        x = self.grid[0]
        y = self.grid[1]
        z = self.grid[2]
        
        if interp3d:
            
            interp = Interp3D(self.data.astype('double', order="C"), x,y,z)
            
        else:   
            interp = RegularGridInterpolator((x,y,z),self.data)
            
        return interp

class Field:
    '''
    The Field class is used to create electric and weighting field objects. 
    We use cartesian coordinates and define each field component as its own 3-dimensional array (fieldx, fieldy, fieldz). 
    grid is the same as in the Potential class. 
    '''
    
    def __init__(self, _name, _fieldx=None, _fieldy=None,_fieldz=None, _grid=None):
        self.name = _name
        self.fieldx = _fieldx
        self.fieldy = _fieldy
        self.fieldz = _fieldz
        self.grid = _grid

    def __str__(self):
        return ("Nessie Field object \n Name: %s \n Size: %s \n"
                % (self.name, np.shape(self.fieldx)))
     
    def interpolate(self, interp3d=False):
        '''
        Returns linear interpolations of the field. 
        interp3d is a cython implementation of grid interpolation, we have modified this from https://github.com/jglaser/interp3d to work for non-regular grids. 
        If interp3d is set to False instead the scipy RegularGridInterpolator will be used. 
        This is a pure python interpolator and is significantly slower. 
        Each interpolator takes input differently so which is used is tracked throughout the simulation.  
        '''
        fieldShape = np.shape(self.fieldx)
        x = self.grid[0]
        y = self.grid[1]
        z = self.grid[2]
        
        if interp3d:
            
            fieldx_interp = Interp3D(self.fieldx.astype('double', order="C"), x,y,z)
            fieldy_interp = Interp3D(self.fieldy.astype('double',order="C"), x,y,z)
            fieldz_interp = Interp3D(self.fieldz.astype('double',order="C"), x,y,z)
            
            fieldMag = np.sqrt(self.fieldx**2+self.fieldy**2+self.fieldz**2)
            fieldMag_interp = Interp3D(fieldMag.astype('double',order="C"), x,y,z)
            
        else:   
            fieldx_interp = RegularGridInterpolator((x,y,z),self.fieldx)
            fieldy_interp = RegularGridInterpolator((x,y,z),self.fieldy)
            fieldz_interp = RegularGridInterpolator((x,y,z),self.fieldz)
            fieldMag = np.sqrt(self.fieldx**2+self.fieldy**2+self.fieldz**2)
            fieldMag_interp = RegularGridInterpolator((x,y,z),fieldMag)
        
        return fieldx_interp, fieldy_interp, fieldz_interp, fieldMag_interp
    
    def saveToH5(self, filename=None):
        if filename is None: filename = self.name
        f_new = h5py.File(filename + ".hf", "w")
        f_new["Ex"] = self.fieldx
        f_new["Ey"] = self.fieldy
        f_new["Ez"] = self.fieldz
        f_new['gridx'] = self.grid[0]
        f_new['gridy'] = self.grid[1]
        f_new['gridz'] = self.grid[2]
        f_new.close()
 

def potentialFromH5(filename, rotate90=False):
    '''
    Import potential from hdf5 file. 
    When saving files to hdf5 from Julia a rotation is applied to the array, so when importing from SSD we need to rotate back.
    '''
    import h5py
    f = h5py.File(filename, 'r')
    data = np.array(f["wp"])
    if rotate90: data = np.rot90(data,axes=(0,2))
    gridx = np.array(f["gridx"])
    gridy = np.array(f["gridy"])
    gridz = np.array(f["gridz"])
    grid = [gridx,gridy,gridz]
    name = filename[:-3]
    potential = Potential(name, data, grid)
    return potential

#Import field from hdf5 file.
def fieldFromH5(filename, rotate90=False):
    '''
    Import field from hdf5 file. 
    When saving files to hdf5 from Julia a rotation is applied to the array, so when importing from SSD we need to rotate back.
    '''
    import h5py
    f = h5py.File(filename, 'r')
    
    if rotate90:
        Ex = np.rot90(np.array(f["Ex"]),axes=(0,2))
        Ey = np.rot90(np.array(f["Ey"]),axes=(0,2))
        Ez = np.rot90(np.array(f["Ez"]),axes=(0,2))
    else:
        Ex = np.array(f["Ex"])
        Ey = np.array(f["Ey"])
        Ez = np.array(f["Ez"])
        
    gridx = np.array(f["gridx"])
    gridy = np.array(f["gridy"])
    gridz = np.array(f["gridz"])
    grid = [gridx,gridy,gridz]
    name = filename[:-3]

    field = Field(name, Ex, Ey,Ez, grid)
    
    return field

def analytical_potential(name, P, bounds,dx,dy,dz):
    '''
    Create a potential using an analytical function of the form lambda x,y,x: f(x,y,z).
    You can then convert this to a field in the normal way to get analytical fields. 
    '''
    x = np.arange(bounds[0][0],bounds[0][1],dx)
    y = np.arange(bounds[1][0],bounds[1][1],dy)
    z = np.arange(bounds[2][0],bounds[2][1],dz)
    grid = [x, y, z]
    X,Y,Z = np.meshgrid(x,y,z,indexing='xy')
    data = P(X,Y,Z)
    return Potential(name, data, grid)

def analytical_field(name, fields, bounds,dx,dy,dz):
    '''
    Create a potential using an analytical function of the form lambda x,y,x: f(x,y,z).
    You can then convert this to a field in the normal way to get analytical fields. 
    '''
    x = np.arange(bounds[0][0],bounds[0][1],dx)
    y = np.arange(bounds[1][0],bounds[1][1],dy)
    z = np.arange(bounds[2][0],bounds[2][1],dz)
    grid = [x, y, z]
    X,Y,Z = np.meshgrid(x,y,z,indexing='xy')
    Ex = fields[0](X,Y,Z)
    Ey = fields[1](X,Y,Z)
    Ez = fields[2](X,Y,Z)
    return Field(name, Ex, Ey, Ez, grid)



from .constants import *
from .mobility import *
import numpy as np

def CoulombForce(q1, q2, pos1, pos2, safety = 7e-9):
    r2 = max(safety**2, sum((pos1-pos2)**2))
    Fs = -kE_SI*q1*q2/r2/epsR_Si
    return Fs*(pos2-pos1)/r2**0.5

def getEffectiveCoulombField(objects):
    Ecoul = np.zeros((len(objects), 3))
    for i in range(len(objects)):
        Q = objects[i].q
        F = np.zeros(3)
        for j in range(len(objects)):
            if i != j:
                F+= CoulombForce(Q, objects[j].q, objects[i].pos[-1], objects[j].pos[-1])
            Ecoul[i] = F/Q
    return Ecoul

def insideBoundaryCheck(pos, bounds):
    return ((pos[0] >= bounds[0][0]) & (pos[0] <= bounds[0][1]) & (pos[1] >= bounds[1][0]) & (pos[1] <= bounds[1][1])
            & (pos[2] >= bounds[2][0]) & (pos[2] <= bounds[2][1]))

def updateQuasiParticles(objects, ds, maxdt, Ex_i, Ey_i, Ez_i, E_i, bounds, temp, diffusion=True, coulomb=False,
                         tauTrap = lambda x, y, z : 1e9, NI = lambda x, y, z : 1e16,
                         mobility_e = canali_mobility_e, mobility_h = canali_mobility_h):
    #Vector to save the effective electric field for each particle at the time step
    Eeff = np.zeros((len(objects), 3))

    # If coulomb is enabled, an effective electric field because of all other charges is added (plasma effects)
    if coulomb:
        Eeff += getEffectiveCoulombField(objects)

    for i in range(len(objects)):
        pos = objects[i].pos[-1]
        dt = maxdt

        Eeff[i] += np.array([Ex_i(pos), Ey_i(pos), Ez_i(pos)])

        if objects[i].q < 0:
            mu = mobility_e(temp, NI(*pos), np.linalg.norm(Eeff[i]))
            dv = -mu*Eeff[i]
            # print(mu, dv, Eeff[i])
        else:
            mu = mobility_h(temp, NI(*pos), np.linalg.norm(Eeff[i]))
            dv = mu*Eeff[i]

        if not coulomb:
            dt = min(abs(ds/np.linalg.norm(dv)), maxdt)

        if diffusion:
            dv += ((diffusion_electron(temp) if objects[i].q < 0 else diffusion_hole(temp))/dt)**0.5*np.random.normal(size=3)

        objects[i].addTime(objects[i].time[-1]+dt)
        objects[i].addVel(dv)
        objects[i].addPos(objects[i].pos[-1]+objects[i].vel[-1]*dt)

        #Check if particles are still inside the specified boundaries, otherwise kill them
        objects[i].alive = insideBoundaryCheck(objects[i].pos[-1], bounds)

        #Check whether a particle is captured. If so, the particle is killed. 
        #dt must be substantially smaller than tauTrap in order for Poisson process to work
        p = dt/tauTrap(*(objects[i].pos[-1]))
        r = np.random.random()
        if r < p:
            objects[i].alive = False

    return objects

if __name__ == "__main__":
    x, y, z, t = propagateCarrier(0, 0, 0, 1e-4, lambda a : [0,], lambda a : [0,], lambda a: [-750e2,], lambda a : [-750e2,], [[-1, 1], [-1, 1], [0, 0.002]], 130, diffusion=True)
